import cv2
from cv2.typing import MatLike


def contour_to_poly(contour):
    """Approximate the contour to a polygon."""
    epsilon = 0.04 * cv2.arcLength(contour, True)  # Adjust epsilon as needed
    approx_polygon = cv2.approxPolyDP(contour, epsilon, True)
    return approx_polygon


def contours_to_poly(contours):
    """Approximate the contours to polygones"""
    return [contour_to_poly(contour) for contour in contours]


def contour_approx_poly_filter_quadrilater(contours):
    return filter_4_edges_contours(contours_to_poly(contours))
    # convex_contours = [cv2.convexHull(contour) for contour in contours]


# See https://docs.opencv.org/4.8.0/d1/d32/tutorial_py_contour_properties.html
def calc_solidity(contour):
    area = cv2.contourArea(contour)
    hull = cv2.convexHull(contour)
    hull_area = cv2.contourArea(hull)
    solidity = float(area) / hull_area
    return solidity


def filter_4_edges_contours(contours):
    return [c for c in contours if len(c) is 4]


def filter_contours_by_size(contours, min_size: int = 2000):
    contour_map = {}

    for i in range(len(contours)):
        if cv2.contourArea(contours[i]) > min_size:
            if calc_solidity(contours[i]) > 0.8:
                contour_map[i] = contours[i]
                continue

    return contour_map


def filter_containing_contours(contour_map, hierarchy):
    """Unefficient algrorithm to filter the second most nested contours."""
    for i in list(contour_map.keys()):
        current_index = i
        while hierarchy[0][current_index][3] > 0:
            if hierarchy[0][current_index][3] in contour_map.keys():
                contour_map.pop(hierarchy[0][current_index][3])
            current_index = hierarchy[0][current_index][3]

    return list(contour_map.values())


## Comparison function for sorting contours
# def get_contour_precedence(contour, cols):
#    # USAGE: final_contour_list.sort(key=lambda x: get_contour_precedence(x, binary.shape[1]))
#    tolerance_factor = 200
#    origin = cv2.boundingRect(contour)
#    return ((origin[1] // tolerance_factor) * tolerance_factor) * cols + origin[0]



from cardscan.pipeline import Pipeline, Transform
import functools
import cv2

from cardscan.image_processing import *


def draw_contours(img, contours):
    if contours is None:
        return img.copy()
    img = copy_bgr(img)
    return [cv2.drawContours(img, contours, -1, (0, 255, 255), 3)]


def debug_contours(contours, parent_pipeline):
    base_img = parent_pipeline.initial_input_getter()
    return draw_contours(base_img, contours)


def debug_contours_def(contours_def, parent_pipeline):
    c, h = contours_def
    img = parent_pipeline.initial_input_getter()
    return draw_contours(img, c)


def debug_contours_map_def(contours_map_def, parent_pipeline):
    cmap, h = contours_map_def
    img = parent_pipeline.initial_input_getter()
    c = list(cmap.values())
    return draw_contours(img, c)


def debug_identity(arg, pipeline):
    return arg


find_contours_def = Transform(
    lambda img, *_: cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE),
    label="Find contours",
    debug=debug_contours_def,
)

filter_contours_shape_size = Transform(
    lambda contours_def, *_: (
        filter_contours_by_size(contours_def[0]),
        contours_def[1],
    ),
    label="Filter contour by size",
    debug=debug_contours_map_def,
)

filter_surronding_contours = Transform(
    lambda contours_def, *_: filter_containing_contours(
        contours_def[0], contours_def[1]
    ),
    label="Filter containing contours",
    debug=debug_contours,
)

approx_filter_quad = Transform(
    lambda contours, _: filter_4_edges_contours(contours_to_poly(contours)),
    label="Approximate quadrilater contours",
    debug=debug_contours,
)

card_contours = Pipeline(
    [
        find_contours_def,
        filter_contours_shape_size,
        filter_surronding_contours,
        approx_filter_quad,
    ],
    label="quadrilater black frame contours pipeline",
)

canny = functools.partial(cv2.Canny, threshold1=150, threshold2=255, apertureSize=3)

card_images = Pipeline(
    [
        Transform(
            lambda img, *_: copy_gray(img),
            label="Gray",
            debug=lambda img, *_: [img],
        ),
        Transform(
            lambda img, *_: canny(img),
            label="Canny",
            debug=lambda arg, *_: [arg],
        ),
        card_contours,
        Transform(
            lambda contours, pipeline, *_: [
                perspective_crop(contour, pipeline.initial_input_getter())
                for contour in contours
            ],
            label="Perspective crop",
            debug=lambda imgs, *_: imgs,
        ),
        Transform(
            lambda imgs, *_: [
                rotate_top_left_corner_low_density(img, corner_size=50) for img in imgs
            ],
            label="Rotate top left corner low density",
            debug=lambda imgs, *_: imgs,
        ),
    ],
    label="Find polygone border pipeline",
)

scan = card_images.run



from typing import Callable, List, Optional, Any, Type

from cv2.typing import MatLike


class Transform:
    def __init__(
        self,
        run: Callable,
        debug: Callable = None,
        label: Optional[str] = "Pipecell",
    ):
        self._run = run
        self.label = label
        self.debug = debug
        self.parent_pipeline = None

    def run_debug(self, input):
        output = self._run(input, self.parent_pipeline)
        if not self.debug:
            debug_cb = None
        debug_cb = lambda: self.debug(output, self.parent_pipeline)
        intermediate_results = {self.label: (output, debug_cb)}
        return output, intermediate_results


class Pipeline:
    initial_input_getter = lambda: None

    def __init__(
        self,
        pipecells: List[Transform | Type["Pipeline"]],
        label: Optional[str] = "Pipeline",
        parent_pipeline=None,
        results: Optional[List[Transform]] = None,
    ):
        for cell in pipecells:
            cell.parent_pipeline = self
        self.pipecells = pipecells
        self.label = label
        self.parent_pipeline = parent_pipeline

    def _run(
        self,
        input: Any,
        parent_pipeline=None,
        results: Optional[List[Transform] | None] = None,
        aggregated_results: Optional[List[Any]] = None,
    ):
        self.initial_input_getter = lambda: input
        input_copy = input

        if results is None:
            for transform in self.pipecells:
                input_copy = transform._run(input_copy, transform.parent_pipeline)
            return input_copy

        for transform in self.pipecells:
            if isinstance(transform, Pipeline):
                input_copy = transform._run(
                    input_copy,
                    transform.parent_pipeline,
                    results=results,
                    aggregated_results=aggregated_results,
                )
            else:
                input_copy = transform._run(input_copy, transform.parent_pipeline)
            if transform in results:
                aggregated_results[results.index(transform)] = input_copy
        if self in results:
            aggregated_results[results.index(self)] = input_copy
        return input_copy

    def run(
        self,
        input: Any,
        parent_pipeline=None,
        results: Optional[List[Transform] | None] = None,
    ):
        if results is None:
            output = self._run(input, parent_pipeline)
        else:
            output = [[]] * len(results)
            self._run(input, parent_pipeline, results, output)
        return output

    def run_debug(self, input: Any):  # TODO typing
        org_input = input.copy()
        self.initial_input_getter = lambda: org_input
        intermediate_results = {}
        for pipecell in self.pipecells:
            input, cell_debug = pipecell.run_debug(input)
            intermediate_results.update(cell_debug)
        return input, intermediate_results



import pygame
import sys
import cv2
import math
from cardscan.pipeline import Pipeline
from cardscan.image_processing import *
from typing import Tuple


def grid_layout_tile_size(tile_ratio, tile_count, window_size):
    window_w, window_h = window_size
    for cols in range(1, 30):
        new_tile_width = window_w // cols
        new_tile_height = int(new_tile_width / tile_ratio)
        rows = math.ceil(tile_count / cols)
        if rows * new_tile_height <= window_h:
            return (new_tile_width, new_tile_height), (rows, cols)


def cv2pygame(frame):
    """Create a pygame surface from opencv BVR or GRAY image. Applies RGB conversion and swap columns and rows."""
    return pygame.surfarray.make_surface(cv2.transpose(copy_rgb(frame)))


def scaleWidthKeepRatio(frame: MatLike, bounding_size: Tuple[int, int]) -> MatLike:
    bound_w, bound_h = bounding_size
    frame_h, frame_w, *_ = frame.shape
    scale_factor_fit_w = bound_w / frame_w
    scale_factor_fit_h = bound_h / frame_h
    scale_factor = min(scale_factor_fit_w, scale_factor_fit_h)
    resized_frame = cv2.resize(
        frame, (int(frame_w * scale_factor), int(frame_h * scale_factor))
    )
    return resized_frame


class Thumbnail:
    def __init__(self, cv2_img, resize, top_left=(0, 0), label: str = "image"):
        img = scaleWidthKeepRatio(cv2_img, resize)
        self.surface = cv2pygame(img)  # pygame.Surface
        self.rect = self.surface.get_rect()
        self.rect.topleft = top_left  # Set the initial position
        self.label = label

    def get_labeled_surface(self):
        # Create a Pygame font object
        font = pygame.font.Font(None, 24)  # None for default font, 36 for font size

        # Define the text and text color
        text_color = (255, 255, 255)  # White color (RGB)
        text_surface = font.render(self.label, True, text_color)
        text_rect = text_surface.get_rect()
        text_rect.topleft = (20, 20)  # self.rect.topleft

        # Draw the text onto the screen
        self.surface.blit(text_surface, text_rect)

        return self.surface


def get_top_left(i: int, cols: int, size: Tuple[int, int]):
    col_i = i % cols
    row_i = i // cols
    return col_i * size[0], row_i * size[1]


def run_app(pipeline: Pipeline):
    pygame.init()

    # Get the current screen resolution (optional)
    screen_info = pygame.display.Info()
    (WINDOW_SIZE) = screen_info.current_w, screen_info.current_h

    BACKGROUND_COLOR = (255, 255, 255)

    # Create a Pygame window
    screen = pygame.display.set_mode(WINDOW_SIZE, pygame.FULLSCREEN)
    pygame.display.set_caption("Image Viewer")

    cam = cv2.VideoCapture(0)

    frame_width = int(cam.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cam.get(cv2.CAP_PROP_FRAME_HEIGHT))

    aspect_ratio = frame_width / frame_height

    selected_index = None
    running = True
    images = []

    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1:  # Left mouse button
                    if selected_index is None:
                        for i, image in enumerate(images):
                            if image.rect.collidepoint(event.pos):
                                selected_index = i
                    else:
                        selected_index = None

        ret, cam_frame = cam.read()
        if ret is None:
            break
        output, intermediate_results = pipeline.run_debug(cam_frame)
        final_frames = []
        for label, (_, debug_cb) in intermediate_results.items():
            if debug_cb is not None:
                final_frames += [(label, frame) for frame in debug_cb()]

        thumbnail_size, (tot_rows, tot_cols) = grid_layout_tile_size(
            aspect_ratio, len(final_frames), WINDOW_SIZE
        )

        images = [
            Thumbnail(
                frame,
                thumbnail_size,
                get_top_left(i, tot_cols, thumbnail_size),
                label=label,
            )
            for i, (label, frame) in enumerate(final_frames)
        ]

        screen.fill(BACKGROUND_COLOR)

        [
            screen.blit(image.get_labeled_surface(), image.rect.topleft)
            for image in images
        ]

        pygame.display.flip()

    cam.release()
    pygame.quit()
    sys.exit()
